import logging 
from typing import Optional

def get_handler(logger: logging.Logger, name: str):
    """
    Get a handler by name from the logger.
    """
    for handler in logger.handlers:
        if not hasattr(handler, "name"):
            continue
        if handler.name == name:
            return handler
    return None

class MemoryHandler(logging.StreamHandler):
    """
    A handler class which buffers logging records in memory, periodically
    flushing them to a target handler. Flushing occurs whenever the buffer
    is full, or when an event of a certain severity or greater is seen.

    Almost a copy of the standard library MemoryHandler, except that it
    has a formatter.
    """
    def __init__(self, name: str,
                 capacity: int, 
                 flushLevel: int = logging.CRITICAL, 
                 target: callable = None,
                 format: Optional[str] = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                 flushOnClose: Optional[bool] = True):
        """
        Initialize the handler with the buffer size, the level at which
        flushing should occur and an optional target.

        Note that without a target being set either here or via setTarget(),
        a MemoryStreamHandler is no use to anyone!

        The ``flushOnClose`` argument is ``True`` for backward compatibility
        reasons - the old behaviour is that when the handler is closed, the
        buffer is flushed, even if the flush level hasn't been exceeded nor the
        capacity exceeded. To prevent this, set ``flushOnClose`` to ``False``.
        """
        logging.StreamHandler.__init__(self)
        self.name = name
        self.capacity = capacity
        self.buffer = []
        self.flushLevel = flushLevel
        self.target = target
        self.flushOnClose = flushOnClose
        self.setFormatter(logging.Formatter(format))

    def value(self):
        """
        Get the buffered messages as a list of strings.
        """
        return "\n".join(self.buffer)
    
    def emit(self, record):
        """
        Emit a record.

        Append the record. If shouldFlush() tells us to, call flush() to process
        the buffer.
        """
        self.buffer.append(self.format(record))
        if self.shouldFlush(record):
            self.flush()

    def shouldFlush(self, record):
        """
        Check for buffer full or a record at the flushLevel or higher.
        """
        return (len(self.buffer) >= self.capacity) or \
                (record.levelno >= self.flushLevel)

    def setTarget(self, target: callable):
        """
        Set the target handler for this handler.
        """
        self.acquire()
        try:
            self.target = target
        finally:
            self.release()

    def flush(self):
        """
        For a MemoryHandler, flushing means just sending the buffered
        records to the target, if there is one. Override if you want
        different behaviour.

        The record buffer is also cleared by this operation.
        """
        self.acquire()
        try:
            if self.target:
                self.target(self.buffer)
                self.buffer.clear()
        finally:
            self.release()

    def close(self, flush: bool = True):
        """
        Flush, if appropriately configured, set the target to None and lose the
        buffer.
        """
        try:
            if self.flushOnClose and flush:
                self.flush()
            self.buffer.clear()
        finally:
            self.acquire()
            try:
                self.target = None
                logging.Handler.close(self)
            finally:
                self.release()



from typing import Optional

def inherit_docstring_from(parent, header: Optional[str] = None):
    """Decorator to inherit docstring from another function

    For example:
    ```python
    @inherit_docstring_from(parent_function)
    def my_function():
        pass
    ```
    
    """
    def decorator(func):
        if header:
            func.__doc__ += "\n" + header + "\n--------------------\n"
        func.__doc__ += parent.__doc__
        return func
    return decorator


from typing import Callable, Optional, Any, List
try:
    from rich.progress import Progress
except:
    Progress = None

class BatchFetch():
    """A helper class to use the paginate API of the Hectiq Console.
    
    It can be used as an iterator or its `get_all` method can be called to get all the results.

    For example:

    ```python
    fetcher = BatchFetch(method=api.get_jobs, batch=10, limit=1000)
    results = fetcher.get_all()
    ```

    or 

    ```python
    fetcher = BatchFetch(method=api.get_jobs, batch=10, limit=1000)
    for results in fetcher:
        print(results)
    ```
    """
    def __init__(self, method: Callable, batch: int = 100, limit: Optional[int] = None, **kwargs):
        """Initialize the BatchFetch object.

        Args:
            method (Callable): The method to call to fetch the data.
            batch (int, optional): The batch size. Defaults to 100.
            limit (Optional[int], optional): The limit of the number of results to fetch. Defaults to None.
            **kwargs: The keyword arguments to pass to the method.
        """

        self.kwargs = kwargs
        self.limit = limit
        self.batch = batch
        self.page = 1
        self.method = method
        self.total_pages = None

    def get_all(self) -> List[Any]:
        """Get all the results. It iterates over the batches and returns the list of results.

        Returns:
            list: The list of results.
        """
        def iterate(progress=None, task=None):
            results = []
            for results_batch in self:
                results += results_batch
                if progress is not None and task is not None:
                    progress.update(task, advance=1, total=self.total_pages)
                if self.limit and len(results) >= self.limit:
                    break
            return results
        if Progress:
            with Progress() as progress:
                task = progress.add_task("[blue]Fetching...")
                return iterate(progress, task)
        else:
            return iterate()

    def get(self) -> List[Any]:
        """Get the next batch of results.
        """
        results = self.method(page=self.page, limit=self.batch, **self.kwargs)
        self.total_pages = results.get("total_pages")
        if results.get("total_pages") < self.page:
            raise StopIteration
        if self.limit and self.page * self.batch > self.limit:
            raise StopIteration
        self.page += 1
        return results["results"]
    
    def __iter__(self):
        return self

    def __next__(self):
        return self.get()
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.page = 1
        pass


import json
import time
import re
import requests
from typing import Optional, List, Dict
from starlette.requests import Request
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.background import BackgroundTask

from hectiq_console import CONSOLE_APP_URL

def response_callback(ressource: str,
                 path: Optional[str] = None,
                 latency: Optional[float] = None,
                 metrics: Optional[dict] = None,
                 annotations: Optional[List[dict]] = None):
    """This method is executed by the HectiqConsoleStarletteMiddleware when the response is sent to the client.
    You should not call this method directly. Instead, use the store_metrics and store_annotation methods.
    """
    headers = {
    }
    body = {
        "path": path,
        "latency": latency,
        "metrics": []
    }
    if metrics is not None:
        body["metrics"] = [{"name": key, "value": value} for key, value in metrics.items()]
    
    if annotations is not None:
        for annot in annotations:
            annot.get("metadata", {})["request_latency"] = latency
            annot.get("metadata", {})["request_path"] = path
    body["annotations"] = annotations

    url = f"{CONSOLE_APP_URL}/app/sender-client/{ressource}/on-request-completed"
    try:
        requests.post(url, data=json.dumps(body, default=str), headers=headers)
    except:
        # If the request fails, we do not want to crash the application
        pass

def send_heartbeat(ressource: str):
    """This method is executed by the HectiqConsoleStarletteMiddleware.
    You should not call this method directly.
    Instead, use the store_metrics and store_annotation methods.
    """
    headers = {
        "content-type": 'application/x-www-form-urlencoded'
    }
    body = {
    }
    url = f"{CONSOLE_APP_URL}/app/sender-client/{ressource}/heartbeat"
    try:
        res = requests.post(url, headers=headers, json=body)
        if res.status_code == 200:
            print(f"✅ Heartbeat succesful with the hectiq console. Your ressource {ressource} is monitored.")
        else:
            print(f"❌ Heartbeat failed with the hectiq console. Your ressource {ressource} is not monitored.")
    except:
        # If the request fails, we do not want to crash the application
        pass

class HectiqConsoleStarletteMiddleware(BaseHTTPMiddleware):
    """
    Middleware that sends metrics to the hectiq console.

    Arguments:
        ressource: The ressource name
        secret_key: The secret key of the ressource
        include_paths: A list of routes to include in the metrics. If None, all routes are included (except the ones in exclude_paths). 
            If specified, only the routes in include_paths are included.
        exclude_paths: A list of routes to exclude from the metrics. If None, all routes are included.
            Example: exclude_paths=["/docs", "/openapi.json"]. You can also use wildcard, for example:
            exclude_paths=["/docs", r"/app/ressources/*/metrics"]
        default_config: A dict of default configuration. 
    """
    def __init__(self, app, 
                 ressource: str, 
                 include_paths: Optional[List] = None,
                 exclude_paths: Optional[List] = None,
                 config: Optional[dict] = {},
                 skip_heartbeat: Optional[bool] = False):
        super().__init__(app)
        self._ressource = ressource
        self.include_paths = include_paths
        self.exclude_paths = exclude_paths
        self.config = config or {}
        if not skip_heartbeat:
            send_heartbeat(ressource=ressource)

    async def dispatch(self, request: Request, call_next):
        path = request.url.path
        request.state.hconsole_config = {k:v for k,v in self.config.items()}
        request.state.hconsole_metrics = {}
        request.state.hconsole_annotations = []

        # Include paths
        if self.include_paths is not None:
            for include_path in self.include_paths:
                if include_path == path or re.match(include_path, path):
                    break
            else:
                return await call_next(request)
            
        # Exclude paths
        if self.exclude_paths is not None:
            for exclude_path in self.exclude_paths:
                if exclude_path == path or re.match(exclude_path, path):
                    return await call_next(request)
        
        start_time = time.time()
        response = await call_next(request)
        process_time = time.time() - start_time

        if request.state.hconsole_config.get("latency_alias"):
            key = request.state.hconsole_config.get("latency_alias")
            if isinstance(key, str):
                request.state.hconsole_metrics[key] = process_time
            elif isinstance(key, list):
                for k in key:
                    request.state.hconsole_metrics[k] = process_time
                    
        if request.state.hconsole_config.get("call_alias"):
            key = request.state.hconsole_config.get("call_alias")
            if isinstance(key, str):
                request.state.hconsole_metrics[key] = 1
            elif isinstance(key, list):
                for k in key:
                    request.state.hconsole_metrics[k] = 1

        response.background = BackgroundTask(response_callback, 
                                             latency=process_time, 
                                             metrics=request.state.hconsole_metrics, 
                                             annotations=request.state.hconsole_annotations, 
                                             path=request.url.path,
                                             ressource=self._ressource)
        return response
    
def store_metrics(request: Request, key: str, value: float):
    """
    Store a metrics in the state of the request
    Should use ContextVar in future releases (remove request params).
    """
    if hasattr(request.state, "hconsole_metrics"):
        request.state.hconsole_metrics[key] = value

def update_config(request: Request, key: str, value: any):
    """
    Update the config in the state of the request.
    """
    if hasattr(request.state, "hconsole_config"):
        request.state.hconsole_config[key] = value

def store_annotation(request: Request, 
                     id: Optional[str] = None, 
                    inputs: Optional[Dict] = None,
                    outputs: Optional[Dict] = None,
                    metadata: Optional[Dict] = None,
                    labels: Optional[List[str]] = None):
    """Store an annotation.

    Params:
        request: The request object (Starlette)
        id: The id of the annotation. If None, the application will assign an unique id.
            We enforce that ids are unique for a given ressource. If you assign the id, then
            you may refer to this annotation id and ressource for updates.
        inputs: Any inputs that you want to store with the annotation. This field is typically the 
            input payload for the model. It must be JSON serializable.
        outputs: Any outputs that you want to store with the annotation. This field is typically the 
            prediction of the model. It must be JSON serializable.
        metadata: Any metadata that you want to store with the annotation. This field is typically the 
            metadata of the model or of the dataset. It must be JSON serializable.
        labels: A list of labels for the annotation. This field is typically the labels of the model.
    """
    if hasattr(request.state, "hconsole_annotations"):
        annotation = {
            "id": id,
            "inputs": inputs,
            "outputs": outputs,
            "metadata": metadata,
            "labels": labels
        }
        request.state.hconsole_annotations.append(annotation)
    


import click


@click.group()
def auth_group():
    pass

@auth_group.command("authenticate")
def cli_authenticate():
    """Authenticate to the Hectiq Console."""
    from hectiq_console.functional import authenticate
    from hectiq_console import CONSOLE_APP_URL
    from pathlib import Path
    import requests
    import os
    import toml
    is_logged = authenticate()
    
    if is_logged:
        # Ask if the user wants to add a new key
        click.secho("You are already logged in.", fg="green")
        should_continue = click.prompt("Do you still want to continue and create a new API key?",
                                        default="y", 
                                        show_default=True, 
                                        type=click.Choice(["y", "n"]))
        if should_continue=="n":
            return
    
    email = click.prompt("Email address", type=str)
    password = click.prompt("Password", type=str, hide_input=True)
    try:
        import socket
        name = socket.gethostname()
    except:
        name = "[unknown hostname]"
    
    name = click.prompt("Alias for the API key:", type=str, default=name)

    # Get the organizations
    res = requests.post(f"{CONSOLE_APP_URL}/app/auth/login", 
                 json={"email": email, "password": password})
    if res.status_code!=200:
        click.echo("Authentication failed.")
        return
    
    user = res.json()
    if user.get("organizations") is None:
        click.echo("You are not part of any organization. Please contact your administrator.")
        return
    
    click.echo("Select an organization:")
    for i, org in enumerate(user.get("organizations")):
        click.echo(f"[{i}] - {org.get('name')}")
    index = click.prompt("Enter the index of the organization:", 
                 type=click.Choice([str(i) for i in range(len(user.get("organizations")))]))
    organization_slug = user.get("organizations")[int(index)]["slug"]

    res = requests.post(f"{CONSOLE_APP_URL}/app/auth/api-keys", 
                  json={"email": email, "password": password, "name": name, "organization": organization_slug})
    
    if res.status_code!=200:
        click.echo("Authentication failed.")
        return
    
    api_key = res.json()
    path = os.getenv("HECTIQ_CONSOLE_CREDENTIALS_FILE", 
                         os.path.join(Path.home(),".hectiq-console", "credentials.toml"))
    
    os.makedirs(os.path.dirname(path), exist_ok=True)
    # Save the key in .hectiq-console/credentials
    with open(path, "a") as f:
        # Dump as TOML
        data = {}
        data[organization_slug] = api_key
        toml.dump(data, f)


    click.echo(f"A new API key has been added to {path}.")
    click.secho("You are now logged in.", fg="green")


from hectiq_console import CONSOLE_APP_URL

import os
import contextvars 
import requests
import time
import datetime as dt
import traceback
from typing import Optional, Union, List, Dict, Literal
from contextlib import contextmanager
from hectiq_console.utils.batch_fetch import BatchFetch
from hectiq_console.utils.log_handler import MemoryHandler, get_handler
from hectiq_console.utils.docstring import inherit_docstring_from

import logging
logger = logging.getLogger()

resource_cvar = contextvars.ContextVar("resource_id", default=None) 
api_key_cvar = contextvars.ContextVar("api_key", default=None) 
organization_cvar = contextvars.ContextVar("organization", default=None)
current_job_cvar = contextvars.ContextVar("current_job", default=None)
current_job_step_cvar = contextvars.ContextVar("current_job_step", default=None)
batch_annotations_cvar = contextvars.ContextVar("batch_annotations", default=None)

def set_ressource(resource: str):
    """Will be removed in the future. Use `set_resource` instead."""
    resource_cvar.set(resource)

def set_resource(resource: str):
    resource_cvar.set(resource)

def set_organization(organization: str):
    organization_cvar.set(organization)

def set_job(job: str):
    current_job_cvar.set(job)

def get_job() -> Optional[str]:
    return current_job_cvar.get()

def set_job_step(job_step: str):
    current_job_step_cvar.set(job_step)

def get_job_step() -> Optional[str]:
    return current_job_step_cvar.get()

def get_organization() -> Optional[str]:
    """Get the organization in the context.

    Returns:
        str: Organization
    """
    return organization_cvar.get()

def get_resource(resource: Optional[str] = None) -> Optional[str]:
    """Get a resource in the context.

    Args:
        resource (Optional[str], optional): Ressource ID of the resource to get. Defaults to None.

    Returns:
        dict: Ressource object
    """
    if resource is None:
        resource = resource_cvar.get()
    if resource is None:
        raise ValueError("You must provide a resource ID to the get_resource method or use `set_resource`.")
    return resource

def authenticate(api_key: Optional[str] = None,
                 organization: Optional[str] = None) -> bool:
    """Authenticate to the Hectiq Console.
    If api_key is None, the api_key is taken from:
     - the environment variable `HECTIQ_CONSOLE_API_KEY`.
     - the api_key located in the file `~/.hectiq-console/credentials.toml` (or the HECTIQ_CONSOLE_CREDENTIALS_FILE).

    Args:
        api_key (Optional[str], optional): API key. Defaults to None.
        organization (Optional[str], optional): Organization. Defaults to None. If None, the organization is taken from the context (set with `set_organization`). If None and no organization is set in the context, the first organization is used. If None and no organization is set in the context and the credentials file contains only one organization, this organization is used.

    Returns:
        bool: True if the authentication is successful, False otherwise.
    """
    if api_key_cvar.get() is not None:
        # Already authenticated
        return True
    
    organization = organization or organization_cvar.get()
    
    if api_key is None:
        # Try to get the api_key from the environment variable
        api_key = os.environ.get("HECTIQ_CONSOLE_API_KEY")

    if api_key is None:
        # Try to get the api_key from the credentials file
        import toml
        from pathlib import Path
        path = os.getenv("HECTIQ_CONSOLE_CREDENTIALS_FILE", 
                         os.path.join(Path.home(),".hectiq-console", "credentials.toml"))
        if not os.path.exists(path):
            return False
        with open(path, "r") as path:
            data = toml.load(path)

        if organization is not None:
            data = data.get(organization, {})
        else:
            # Multiple organizations
            if len(data)==0:
                return False
            
            organization = list(data.keys())[0]
            set_organization(organization) # Set the organization in the context
            data = list(data.values())[0]
      
        api_key = data.get("value")

        if api_key is None:
            return False

    api_key_cvar.set(api_key)
    return True

def get_authentification_headers() -> dict:
    """Get the authentification headers for the Hectiq Console. 
    Do not use this method directly. Instead, use the `authenticate` method.

    Returns:
        dict: Headers
    """
    is_logged = authenticate()
    if not is_logged:
        raise ValueError(f"You must authenticate to the Hectiq Console using the `hectiq_console.functional.authenticate` method or the command line `hectiq-console authenticate`. This error may occur if the organization ({get_organization()}) is not set or incorrect.")
    api_key = api_key_cvar.get()
    if api_key is None:
        raise ValueError("You must authenticate to the Hectiq Console using the `authenticate` method.")
    return {"X-API-Key": f"{api_key}"}

def create_incident(title: str, 
                    description: Optional[str] = None,
                    filenames: Optional[List] = None, 
                    resource: Optional[str] = None):
    """Create an incident in the Hectiq Console.

    Args:
        title (str): Title of the incident
        description (Optional[str], optional): Description of the incident. Defaults to None.
        filenames (Optional[List], optional): List of filenames to attach to the incident. Defaults to None.
        resource (Optional[str], optional): Ressource ID of the resource to which the incident is related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    body = {"name": title, "description": description}
    if filenames is not None:
        body["files"] = []
        for filename in filenames:
            assert os.path.exists(filename), f"File {filename} does not exist."
            name = os.path.basename(filename)
            num_bytes = os.path.getsize(filename)
            extension = os.path.splitext(filename)[1].replace(".", "")
            body["files"].append({"name": name, "num_bytes": num_bytes, "extension": extension})
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/incidents", 
                 json=body)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while creating the incident with hectiq_console.create_incident: {res.text}")
        return
    
    # Upload the files
    if filenames is not None:
        from hectiq_console.upload import upload_file
        for filename, policy in zip(filenames, res.json()["policies"]):
            upload_file(filepath=filename, policy=policy)

def add_file(filename: str,
             resource: Optional[str] = None):
    """Add a file to a resource in the Hectiq Console.

    Args:
        filename (str): Name of the file
        resource (Optional[str], optional): Ressource ID of the resource to which the file is related. 
            Defaults to None.
    """
    from hectiq_console.upload import upload_file
    resource = get_resource(resource)

    assert os.path.exists(filename), f"File {filename} does not exist."
    name = os.path.basename(filename)
    num_bytes = os.path.getsize(filename)
    extension = os.path.splitext(filename)[1].replace(".", "")
    json = {"name": name, "num_bytes": num_bytes, "extension": extension}
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/files", json=json)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while creating the file with hectiq_console.add_file: {res.text}")
        return
    try:
        policy = res.json()
    except:
        logger.error(f"⚠️ Error while creating the file with hectiq_console.add_file: {res.text}")
        return
    upload_file(filepath=filename, policy=policy)

def add_metrics(name: str, 
                value: Union[float, int], 
                resource: Optional[str] = None):
    """Add metrics to the Hectiq Console.

    Args:
        key (str): Key of the metrics
        value (Union[float, int]): Value of the metrics
        resource (Optional[str], optional): Ressource ID of the resource to which the metrics are related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    body = {
        "metrics" : [{"name": name, "value": value}]
    }
    requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/metrics", 
                 json=body)
    
def add_annotation(id: Optional[str] = None, 
                    inputs: Optional[Dict] = None,
                    outputs: Optional[Dict] = None,
                    metadata: Optional[Dict] = None,
                    resource: Optional[str] = None):
    """Add an annotation to the Hectiq Console. 
    For multiple annotations, you can use

    ```
    with hectiq_console.batch_annotations():
        hectiq_console.add_annotation(...)
        hectiq_console.add_annotation(...)
    ```
    Args:
        id (Optional[str], optional): ID of the annotation. Defaults to None.
        inputs (Optional[Dict], optional): Inputs of the annotation. Defaults to None.
        outputs (Optional[Dict], optional): Outputs of the annotation. Defaults to None.
        metadata (Optional[Dict], optional): Metadata of the annotation. Defaults to None.
        resource (Optional[str], optional): Ressource ID of the resource to which the annotation is related.
    """
    resource = get_resource(resource)
    data = {"inputs": inputs, "outputs": outputs, "metadata": metadata}
    if id is not None:
        data["id"] = id

    if batch_annotations_cvar.get() is not None:
        val = batch_annotations_cvar.get()
        val.append(data)
        batch_annotations_cvar.set(val)
        return
    
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotations", 
                        json=data)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while creating the annotation with hectiq_console.add_annotation: {res.text}")

def add_annotations(annotations: List[Dict],
                    resource: Optional[str] = None):
    """Add multiple annotations to the Hectiq Console.

    Args:
        annotations (List[Dict]): Annotations with optional keys `id`, `inputs`, `outputs` and `metadata`. 
        resource (Optional[str], optional): Ressource ID of the resource to which the annotations are related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    data = {"annotations": annotations}
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/batch-annotations",
                        json=data)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while creating the annotations with hectiq_console.add_annotations: {res.text}")

def download_annotation(id: str,
                        resource: Optional[str] = None) -> dict:
    """Download an annotation from the Hectiq Console.

    Args:
        id (str): ID of the annotation
        resource (Optional[str], optional): Ressource ID of the resource to which the annotation is related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()
    res = requests.get(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotations/{id}", headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while downloading the annotation with hectiq_console.download_annotation: {res.text}")
        return
    return res.json()

def download_annotations(from_date: Optional[dt.datetime] = None,
                        to_date: Optional[dt.datetime] = None,
                        labels: Optional[List[str]] = None,
                        label_set: Literal["any", "all", "none", "empty"] = "any",
                        fields: Optional[List[str]] = ["private_id", "inputs", "outputs", "metadata"],
                        page: Optional[int] = 1,
                        limit: Optional[int] = 100,
                        order_by: Optional[str] = None,
                        order_direction: Literal["asc", "desc"] = "asc",
                         resource: Optional[str] = None):
    """Download all annotations from the Hectiq Console.

    Args:   
        from_date (Optional[dt.datetime]): Start date of creation of the annotations. Defaults to None.
        to_date (Optional[dt.datetime]): End date of creation of the annotations. Defaults to None.
        labels (Optional[List[str]]): Labels of the annotations. Defaults to None.
        label_set (Literal["any", "all", "none", "empty"]): Set of labels. Defaults to "any".
            - "any": At least one label must be present
            - "all": All labels must be present
            - "none": No label from the label list must be present
            - "empty": No label must be present (label list must be empty)
        fields (Optional[List[str]]): Fields to return. Defaults to ["private_id", "inputs", "outputs", "metadata"].
        page (Optional[int]): Page number. Defaults to 1.
        limit (Optional[int]): Number of annotations per page. Defaults to 100.
        order_by (Optional[str]): Field to order by. Defaults to None.
        order_direction (Literal["asc", "desc"]): Order direction. Defaults to "asc".
        resource (Optional[str]): Ressource ID of the resource to which the annotations are related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()

    params = {
        "fields": fields,
        "created_from_date": from_date.toisoformat() if from_date is not None else None,
        "created_to_date": to_date.toisoformat() if to_date is not None else None,
        "labels": labels,
        "label_set": label_set,
        "page": page,
        "limit": limit,
        "order_by": order_by,
        "order_direction": order_direction
    }
    res = requests.get(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotations", params=params, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while downloading the annotations with hectiq_console.download_annotations: {res.text}")
        return
    return res.json()

def search_annotation_labels(search: Optional[str] = None,
                             fields: Optional[List[str]] = ["id", "label"],
                                page: Optional[int] = 1,
                                limit: Optional[int] = 100,
                                order_by: Optional[str] = None,
                                order_direction: Literal["asc", "desc"] = "asc",
                                resource: Optional[str] = None):
    """Search annotation labels from the Hectiq Console.

    Args:
        search (Optional[str]): Search string. Defaults to None.
        fields (Optional[List[str]]): Fields to return. Defaults to ["id", "label"].
        page (Optional[int]): Page number. Defaults to 1.
        limit (Optional[int]): Number of annotations per page. Defaults to 100.
        order_by (Optional[str]): Field to order by. Defaults to None.
        order_direction (Literal["asc", "desc"]): Order direction. Defaults to "asc".
        resource (Optional[str]): Ressource ID of the resource to which the annotations are related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()

    params = {
        "fields": fields,
        "search": search,
        "page": page,
        "limit": limit,
        "order_by": order_by,
        "order_direction": order_direction
    }
    res = requests.get(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotation-labels", params=params, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while fetching the annotation labels with hectiq_console.search_annotation_labels: {res.text}")
        return
    return res.json()

def count_annotations(from_date: Optional[dt.datetime] = None,
                        to_date: Optional[dt.datetime] = None,
                        labels: Optional[List[str]] = None,
                        label_set: Literal["any", "all", "none", "empty"] = "any",
                        resource: Optional[str] = None):
    """Count annotations from the Hectiq Console.

    Args:
        from_date (Optional[dt.datetime]): Start date of creation of the annotations. Defaults to None.
        to_date (Optional[dt.datetime]): End date of creation of the annotations. Defaults to None.
        labels (Optional[List[str]]): Labels of the annotations. Defaults to None.
        label_set (Literal["any", "all", "none", "empty"]): Set of labels. Defaults to "any".
            - "any": At least one label must be present
            - "all": All labels must be present
            - "none": No label from the label list must be present
            - "empty": No label must be present (label list must be empty)
        resource (Optional[str]): Ressource ID of the resource to which the annotations are related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()

    params = {
        "created_from_date": from_date.toisoformat() if from_date is not None else None,
        "created_to_date": to_date.toisoformat() if to_date is not None else None,
        "labels": labels,
        "label_set": label_set
    }
    res = requests.get(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/count-annotations", params=params, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while counting the annotations with hectiq_console.count_annotations: {res.text}")
        return
    return res.json()["count"]

def attach_annotation_label(annotation_id: str, 
                               label_id: str,
                               resource: Optional[str] = None):
    """Attach a label to an annotation in the Hectiq Console.
    If the label is already attached, nothing is done.

    Args:
        annotation_id (str): ID of the annotation
        label_id (str): ID of the label
        resource (Optional[str], optional): Ressource ID of the resource to which the annotation is related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotations/{annotation_id}/add-label/{label_id}", 
                 headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while attaching the label to the annotation with hectiq_console.attach_label_to_annotation: {res.text}")
        return
    
    return True

def detach_annotation_label(annotation_id: str, 
                               label_id: str,
                               resource: Optional[str] = None):
    """Remove a label to an annotation in the Hectiq Console.
    If the label is not attached, nothing is done.

    Args:
        annotation_id (str): ID of the annotation
        label_id (str): ID of the label
        resource (Optional[str], optional): Ressource ID of the resource to which the annotation is related. 
            Defaults to None.
    """
    resource = get_resource(resource)
    headers = get_authentification_headers()
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/annotations/{annotation_id}/remove-label/{label_id}", 
                 headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while attaching the label to the annotation with hectiq_console.remove_label_to_annotation: {res.text}")
        return
    
    return True

def start_job(name: str, 
              description: Optional[str] = None, 
              metadata: Optional[Dict] = None,
              status: Optional[str] = None,
              log_level: int = logging.INFO,
              logger_name: Optional[str] = None,
              resource: Optional[str] = None):
    """Start a job in the Hectiq Console.
    """

    if get_job() is not None:
        # Current job already started
        return
    
    resource = get_resource()
    body = {"name": name, 
            "description": description, 
            "metadata": metadata,
            "status": status or "running"}
    headers = get_authentification_headers()
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/jobs", 
                 json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while starting the job with hectiq_console.start_job: {res.text}")
        return
    job_id = res.json().get("id")
    set_job(job_id)

    handler = MemoryHandler(name="hectiq-console-job-handler",
                            capacity=100000, # Flush every 5000 lines
                            target=append_current_job_logs)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root = logging.getLogger(logger_name)
    root.addHandler(handler)
    root.setLevel(log_level)
    handler.setLevel(log_level)

    return True

def end_job(status: Optional[str] = None):
    """End the current job.
    """
    job_id = get_job()
    if job_id is None:
        # No job started
        return
    resource = get_resource()
    headers = get_authentification_headers()
    body = {"status": status}

    root = logging.getLogger()
    handler = get_handler(logger=root, name="hectiq-console-job-handler")
    logs = handler.value() if handler is not None else None
    if logs is not None and len(logs)>0:
        if isinstance(logs, list):
            logs = "\n".join(logs)
        body["logs"] = logs
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/jobs/{job_id}/end", json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while ending the job with hectiq_console.end_job: {res.text}")
        return
    set_job(None)
    if handler:
        handler.close(flush=False)

def update_current_job(name: Optional[str] = None,
                        description: Optional[str] = None, 
                        metadata: Optional[Dict] = None,
                        logs: Optional[str] = None,
                        status: Optional[str] = None):
    """Update the current job. 
    Only the non-null values are updated.
    """
    job_id = get_job()
    if job_id is None:
        # No job started
        return
    resource = get_resource()
    body = {"name": name, 
        "description": description, 
        "metadata": metadata,
        "logs": logs,
        "status": status}
    # Remove null values
    body = {k: v for k, v in body.items() if v is not None}
    if len(body)==0:
        return
    headers = get_authentification_headers()
    res = requests.put(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/jobs/{job_id}", 
                json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while updating the job with hectiq_console.update_current_job: {res.text}")
        return
    return True

def append_current_job_logs(logs: Union[str, List[str]]):
    # Format the logs
    if len(logs)==0:
        return
    if isinstance(logs, list):
        logs = "\n".join(logs)
    update_current_job(logs=logs)

def start_job_step(name: str, 
              description: Optional[str] = None, 
              metadata: Optional[Dict] = None,
              status: Optional[str] = None,
              log_level: int = logging.INFO,
              logger_name: Optional[str] = None,
              resource: Optional[str] = None):
    """Start a job in the Hectiq Console.
    """
    job = get_job()
    if job is None:
        # No job started
        return
    
    if get_job_step() is not None:
        # Job step is started, close it
        end_job_step()
    
    resource = get_resource()
    body = {"name": name, 
            "description": description, 
            "metadata": metadata,
            "status": status or "running"}
    headers = get_authentification_headers()
    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/jobs/{job}/steps", 
                 json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while starting the job step with hectiq_console.start_job_step: {res.text}")
        return
    step_id = res.json().get("id")
    set_job_step(step_id)

    handler = MemoryHandler(name="hectiq-console-job-step-handler", 
                            capacity=100000,
                            target=append_current_job_step_logs)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root = logging.getLogger(logger_name)
    root.addHandler(handler)
    root.setLevel(log_level)

    handler.setLevel(log_level)

    return True

def end_job_step(status: Optional[str] = None):
    """End the current job.
    """
    job_id = get_job()
    if job_id is None:
        # No job started
        return
    step_id = get_job_step()
    if step_id is None:
        # No job step started
        return
    resource = get_resource()
    headers = get_authentification_headers()
    handler = get_handler(logger=logging.getLogger(), name="hectiq-console-job-step-handler")
    logs = handler.value() if handler is not None else None
    body = {"status": status}
    if logs is not None and len(logs)>0:
        if isinstance(logs, list):
            logs = "\n".join(logs)
        body["logs"] = logs

    res = requests.post(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/job-steps/{step_id}/end", json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while ending the job step with hectiq_console.end_job_step: {res.text}")
        return
    set_job_step(None)
    if handler:
        handler.close(flush=False)

def update_current_job_step(name: Optional[str] = None,
                          description: Optional[str] = None, 
                          metadata: Optional[Dict] = None,
                          status: Optional[str] = None,
                          logs: Optional[str] = None):
    """Update the current job. 
    Only the non-null values are updated.
    """
    job_id = get_job()
    if job_id is None:
        # No job started
        return
    step_id = get_job_step()
    if step_id is None:
        # No job step started
        return
    resource = get_resource()
    body = {"name": name, 
        "description": description, 
        "metadata": metadata,
        "logs": logs,
        "status": status}
    # Remove null values
    body = {k: v for k, v in body.items() if v is not None}
    if len(body)==0:
        return
    headers = get_authentification_headers()
    res = requests.put(f"{CONSOLE_APP_URL}/app/sender-client/{resource}/job-steps/{step_id}", 
                json=body, headers=headers)
    if res.status_code != 200:
        logger.error(f"⚠️ Error while updating the job step with hectiq_console.update_current_job_step: {res.text}")
        return
    return True

def append_current_job_step_logs(logs: Union[str, List[str]]):
    # Format the logs
    if len(logs)==0:
        return
    if isinstance(logs, list):
        logs = "\n".join(logs)
    update_current_job_step(logs=logs)

@inherit_docstring_from(BatchFetch, header="Below is the BatchFetch (parent) docstring")
class BatchDownloadAnnotations(BatchFetch):
    """A helper class to download annotations. 

    Example:

    ```python
    for annotations in BatchDownloadAnnotations(batch=100):
        print(f"✅ Downloaded {len(annotations)} annotations.")
    ```
    """
    @inherit_docstring_from(BatchFetch.__init__, header="BatchFetch.__init__ docstring.")
    def __init__(self, **kwargs):
        """The **kwargs are passed to the BatchFetch class.
        The method used is `hectiq_console.functional.download_annotations`.
        """
        super().__init__(method=download_annotations, **kwargs)

@inherit_docstring_from(BatchFetch, header="Below is the BatchFetch (parent) docstring")
class BatchSearchAnnotationLabels(BatchFetch):
    """A helper class to search annotation labels. 

    Example:

    ```python
    for labels in BatchSearchAnnotationLabels(batch=100):
        print(f"✅ Downloaded {len(labels)} labels.")
    ```
    """
    @inherit_docstring_from(BatchFetch.__init__, header="BatchFetch.__init__ docstring.")
    def __init__(self, **kwargs):
        """The **kwargs are passed to the BatchFetch class.
        The method used is `hectiq_console.functional.search_annotation_labels`.
        """
        super().__init__(method=search_annotation_labels, **kwargs)

@contextmanager
def timer_context(name: str, 
                  resource: Optional[str] = None):
    """Context manager to time a block of code.

    Args:
        key (str): Key of the timer
        resource (Optional[str], optional): Ressource ID of the resource to which the timer is related. 
            Defaults to None.
    """
    start = time.time()
    yield
    end = time.time()
    duration = end - start
    add_metrics(name=name, value=duration, resource=resource)

@contextmanager
def batch_annotations():
    """Context manager to batch the creation of annotations.
    """

    batch_annotations_cvar.set([])
    yield
    if len(batch_annotations_cvar.get())>0:
        add_annotations(annotations=batch_annotations_cvar.get())
    batch_annotations_cvar.set(None)

@contextmanager
def job_context(name: str, 
                description: Optional[str] = None, 
                metadata: Optional[Dict] = None,
                status: Optional[str] = None,
                log_level: int = logging.INFO,
                logger_name: Optional[str] = None,
                timer_metrics: Optional[str] = None,
                resource: Optional[str] = None):
    """Context manager to start a job.

    Args:
        name (str): Name of the job
        description (Optional[str], optional): Description of the job. Defaults to None.
        metadata (Optional[Dict], optional): Metadata of the job. Defaults to None.
        status (Optional[str], optional): Status of the job. Defaults to None.
        resource (Optional[str], optional): Ressource ID of the resource to which the job is related. 
            Defaults to None.
        log_level (int, optional): Log level. Defaults to logging.INFO. Other possible values are:
            - logging.DEBUG = 10
            - logging.INFO = 20
            - logging.WARNING = 30
            - logging.ERROR = 40
            - logging.CRITICAL = 50

    """
    start_job(name=name, 
              description=description, 
              metadata=metadata, 
              status=status,
              log_level=log_level,
              logger_name=logger_name,
              resource=resource)
    
    exception_raised = False
    resource = get_resource(resource)

    try:
        if timer_metrics:
            with timer_context(name=timer_metrics, resource=resource):
                yield
        else:
            yield
    except Exception as e:
        exception_raised = True
        raise e
    finally:
        # Catch all exceptions to end the job step
        if exception_raised:
            logger.error(traceback.format_exc())
            end_job_step(status="failed")
            end_job(status="failed")
        else:
            end_job_step()
            end_job()
            
@contextmanager
def job_step_context(name: str, 
                      description: Optional[str] = None, 
                      metadata: Optional[Dict] = None,
                      status: Optional[str] = None,
                      log_level: int = logging.INFO,
                      logger_name: Optional[str] = None,
                      timer_metrics: Optional[str] = None,
                      resource: Optional[str] = None):
    """Context manager to start a job step.

    Args:
        name (str): Name of the job step
        description (Optional[str], optional): Description of the job step. Defaults to None.
        metadata (Optional[Dict], optional): Metadata of the job step. Defaults to None.
        status (Optional[str], optional): Status of the job step. Defaults to None.
        resource (Optional[str], optional): Ressource ID of the resource to which the job step is related. 
            Defaults to None.
        log_level (int, optional): Log level. Defaults to logging.INFO. Other possible values are:
            - logging.DEBUG = 10
            - logging.INFO = 20
            - logging.WARNING = 30
            - logging.ERROR = 40
            - logging.CRITICAL = 50
    """
    start_job_step(name=name, 
              description=description, 
              metadata=metadata, 
              status=status,
              log_level=log_level,
              logger_name=logger_name,
              resource=resource)
    exception_raised = False
    try:
        if timer_metrics:
            with timer_context(name=timer_metrics, resource=resource):
                yield
        else:
            yield
    except Exception as e:
        exception_raised = True
        raise e
    finally:
        # Catch all exceptions to end the job step
        if exception_raised:
            logger.error(traceback.format_exc())
            end_job_step(status="failed")
        else:
            end_job_step()